package ch.fritteli.labyrinth.generator;

import ch.fritteli.labyrinth.generator.model.Direction;
import ch.fritteli.labyrinth.generator.model.Labyrinth;
import ch.fritteli.labyrinth.generator.model.Position;
import ch.fritteli.labyrinth.generator.model.Tile;
import io.vavr.control.Option;
import java.util.Deque;
import java.util.LinkedList;
import java.util.Random;
import lombok.NonNull;

public class Generator {

    @NonNull
    private final Labyrinth labyrinth;
    @NonNull
    private final Random random;
    @NonNull
    private final Deque<Position> positions = new LinkedList<>();

    public Generator(@NonNull final Labyrinth labyrinth) {
        this.labyrinth = labyrinth;
        this.random = new Random(labyrinth.getRandomSeed());
    }

    public void run() {
        this.preDig();
        this.dig();
        this.postDig();
    }

    private void preDig() {
        final Position end = this.labyrinth.getEnd();
        final Tile endTile = this.labyrinth.getEndTile();

        if (end.getY() == 0) {
            endTile.enableDiggingToOrFrom(Direction.TOP);
            endTile.digFrom(Direction.TOP);
        } else if (end.getX() == 0) {
            endTile.enableDiggingToOrFrom(Direction.LEFT);
            endTile.digFrom(Direction.LEFT);
        } else if (end.getY() == this.labyrinth.getHeight() - 1) {
            endTile.enableDiggingToOrFrom(Direction.BOTTOM);
            endTile.digFrom(Direction.BOTTOM);
        } else if (end.getX() == this.labyrinth.getWidth() - 1) {
            endTile.enableDiggingToOrFrom(Direction.RIGHT);
            endTile.digFrom(Direction.RIGHT);
        }

        this.positions.push(end);
    }

    private void dig() {
        while (!this.positions.isEmpty()) {
            final Position currentPosition = this.positions.peek();
            final Tile currentTile = this.labyrinth.getTileAt(currentPosition).get();
            final Option<Direction> directionToDigTo = currentTile.getRandomAvailableDirection(this.random);
            if (directionToDigTo.isDefined()) {
                final Direction digTo = directionToDigTo.get();
                final Direction digFrom = digTo.invert();
                final Position neighborPosition = currentPosition.move(digTo);
                final Tile neighborTile = this.labyrinth.getTileAt(neighborPosition).get();
                if (currentTile.digTo(digTo) && neighborTile.digFrom(digFrom)) {
                    // all ok!
                    this.positions.push(neighborPosition);
                    if (neighborPosition.equals(this.labyrinth.getStart())) {
                        this.markSolution();
                    }
                } else {
                    // Hm, didn't work.
                    currentTile.undigTo(digTo);
                    currentTile.preventDiggingToOrFrom(digTo);
                }
            } else {
                this.positions.pop();
            }
        }
    }

    private void markSolution() {
        this.positions.forEach(position -> this.labyrinth.getTileAt(position).get().setSolution());
    }

    private void postDig() {
        final Position start = this.labyrinth.getStart();
        final Tile startTile = this.labyrinth.getStartTile();

        if (start.getY() == 0) {
            startTile.enableDiggingToOrFrom(Direction.TOP);
            startTile.digTo(Direction.TOP);
        } else if (start.getX() == 0) {
            startTile.enableDiggingToOrFrom(Direction.LEFT);
            startTile.digTo(Direction.LEFT);
        } else if (start.getY() == this.labyrinth.getHeight() - 1) {
            startTile.enableDiggingToOrFrom(Direction.BOTTOM);
            startTile.digTo(Direction.BOTTOM);
        } else if (start.getX() == this.labyrinth.getWidth() - 1) {
            startTile.enableDiggingToOrFrom(Direction.RIGHT);
            startTile.digTo(Direction.RIGHT);
        }
    }
}
